(* --------------------------------------------------------------- *)
(** * Basic lemmas about F_sub.                                    *)
(* --------------------------------------------------------------- *)

Set Implicit Arguments.

Require Export F_sub_defs.

(* --------------------------------------------------------------- *)
(** ** Functorial properties of type expressions.                  *)
(* --------------------------------------------------------------- *)

Lemma fct_fct : forall (X : Set)
  (t : T X) (Y Z : Set) (f : X -> Y) (g : Y -> Z),
  fct g (fct f t) = fct (fun x => g (f x)) t.
Proof.
induction t; intros; simpl; auto.
congr. auto.
rewrite IHt2. congr.
extens x; destruct x; reflexivity.
Qed.

Lemma fct_id : forall (X : Set) (f : X -> X) x,
  (forall u, f u = u) -> fct f x = x.
Proof.
induction x; simpl; intros; auto.
congr. congr. auto.
apply IHx2. destruct u; simpl.
rewrite H. reflexivity.
reflexivity.
Qed.

Definition fct_bij X Y (f : Bij X Y) : Bij (T X) (T Y).
intros.
split with (fct f) (fct (bij_bw f)).
abstract (intros; rewrite fct_fct; apply fct_id; apply bij_l).
abstract (intros; rewrite fct_fct; apply fct_id; apply bij_r).
Defined.

Lemma fct_inj : forall (X : Set) (m n : T X) (Y : Set) (f : X -> Y),
  (forall x y, f x = f y -> x = y) ->
  fct f m = fct f n -> m = n.
Proof.
induction m; destruct n; simpl; try discriminate 2; intros; auto.
congr.
inversion H0; clear H0; auto.
inversion H0; clear H0.
repeat congr; firstorder idtac.
inversion H0. repeat congr.
firstorder. clear IHm1.
apply (IHm2 n2 _ (optfct f)); auto.
destruct x; destruct y;
  (first [ discriminate 1 | inversion 1 ]; auto).
Qed.

Lemma fct_eq_eta : forall X (t : T X) (Y : Set) (f : X -> Y) (x : Y),
  fct f t = eta x -> exists y : X, x = f y /\ t = eta y.
Proof.
induction t; inversion 1. subst x0.
exists x; auto.
Qed.

Lemma fct_eq_top : forall X (t : T X) (Y : Set) (f : X -> Y),
  fct f t = top _ -> t = top _.
Proof.
induction t; inversion 1; reflexivity.
Qed.

Lemma fct_eq_arr : forall X (t : T X) (Y : Set) (f : X -> Y)
  (s1 s2 : T Y),
  fct f t = arr s1 s2 ->
  exists t1, exists t2,
    t = arr t1 t2 /\ s1 = fct f t1 /\ s2 = fct f t2.
Proof.
induction t; simpl; try discriminate 1; intros.
inversion H; clear H. subst s1 s2.
exists t1. exists t2. auto.
Qed.

Lemma fct_eq_uni : forall X (t : T X)
  (Y : Set) (f : X -> Y) (s1 : T Y) s2,
  fct f t = uni s1 s2 ->
  exists t1, exists t2,
    t = uni t1 t2 /\ s1 = fct f t1 /\ s2 = fct (optfct f) t2.
Proof.
induction t; simpl; try discriminate 1; intros.
inversion H; clear H. subst s1 s2.
exists t1. exists t2. auto.
Qed.

Lemma shift_eq_eta : forall X (t : T X) (x : option X),
  shift t = eta x -> exists y : X, x = Some y /\ t = eta y.
Proof.
induction t; inversion 1. subst x0.
exists x. auto.
Qed.

Lemma shift_eq_top : forall X (t : T X), shift t = top _ -> t = top _.
Proof.
induction t; inversion 1; reflexivity.
Qed.

Lemma shift_eq_arr : forall X (t : T X) (s1 s2 : T (option X)),
  shift t = arr s1 s2 ->
  exists t1, exists t2,
    t = arr t1 t2 /\ s1 = shift t1 /\ s2 = shift t2.
Proof.
induction t; try discriminate 1; intros.
change (arr (shift t1) (shift t2) = arr s1 s2) in H.
inversion H; clear H. subst s1 s2.
exists t1. exists t2. auto.
Qed.

Lemma shift_eq_uni : forall X (t : T X) (s1 : T (option X)) s2,
  shift t = uni s1 s2 ->
  exists t1, exists t2,
    t = uni t1 t2 /\ s1 = shift t1 /\ s2 = fct (optfct (@Some _)) t2.
Proof.
induction t; try discriminate 1; intros.
inversion H; clear H. subst s1 s2.
exists t1. exists t2. auto.
Qed.

Section Types_FProd.

Variable (X1 X2 B U : Set).
Variable (f1 : X1 -> B) (f2 : X2 -> B).
Variable (p1 : U -> X1) (p2 : U -> X2).
Variable FP : FProd f1 f2 p1 p2.

Let sig_h : forall s1 s2 (H : fct f1 s1 = fct f2 s2),
  { u : T U | fct p1 u = s1 /\ fct p2 u = s2 }.
Proof.
intros.
generalize X1 s1 X2 s2 B U f1 f2 p1 p2 FP H.
clear X1 s1 X2 s2 B U f1 f2 p1 p2 FP H.
induction s1.

(* -4: s1 = eta x *)
destruct s2; simpl; try discriminate 2. intros.
assert (Hx : f1 x = f2 x0). inversion H. reflexivity.
split with (eta (fpair FP x x0 Hx)).
simpl. rewrite fpair_1. rewrite fpair_2. auto.

(* -3: s1 = top *)
destruct s2; simpl; try discriminate 2. intros.
split with (top U). simpl. tauto.

(* -2: s1 = arr *)
destruct s2; simpl; try discriminate 2. intros.
inversion H; clear H.
destruct (IHs1_1 X2 s2_1 B U f1 f2 p1 p2 FP H1) as [a [Ha1 Ha2]].
destruct (IHs1_2 X2 s2_2 B U f1 f2 p1 p2 FP H2) as [b [Hb1 Hb2]].
split with (arr a b). simpl.
rewrite Ha1. rewrite Ha2. rewrite Hb1. rewrite Hb2. tauto.

(* -2: s1 = uni *)
destruct s2; simpl; try discriminate 2. intros.
inversion H; clear H.
destruct (IHs1_1 X2 s2_1 B U f1 f2 p1 p2 FP H1) as [a [Ha1 Ha2]].
destruct (IHs1_2 (option X2) s2_2 (option B) (option U)
           (optfct f1) (optfct f2) (optfct p1) (optfct p2)
           (FProd_option FP) H2)
  as [b [Hb1 Hb2]].
split with (uni a b). simpl.
rewrite Ha1. rewrite Ha2. rewrite Hb1. rewrite Hb2. tauto.
Qed.

Let h x1 x2 (H : fct f1 x1 = fct f2 x2) : T U :=
  let (u, _) := sig_h x1 x2 H in u.

Lemma unique : forall u v,
  fct p1 u = fct p1 v -> fct p2 u = fct p2 v -> u = v.
Proof.
clear sig_h h.
intros.
generalize u v X1 X2 B f1 f2 p1 p2 FP H H0.
clear u v X1 X2 B f1 f2 p1 p2 FP H H0.

induction u.

destruct v; simpl; try discriminate 2. intros. congr.
apply (fprod_unique FP).
inversion H. reflexivity.
inversion H0. reflexivity.

destruct v; simpl; try discriminate 2. reflexivity.

destruct v; simpl; try discriminate 2. intros.
inversion H; clear H. inversion H0; clear H0.
congr. congr. eauto. eauto.

destruct v; simpl; try discriminate 2. intros.
inversion H; clear H. inversion H0; clear H0.
congr. congr.
eauto.
apply (IHu2 v2 _ _ _ (optfct f1) (optfct f2)
 (optfct p1) (optfct p2)).
apply FProd_option.
assumption.
assumption.
auto.
Qed.

Definition FProd_T : FProd (fct f1) (fct f2) (fct p1) (fct p2).
Proof.
split with h; intros; unfold h.
do 2 rewrite fct_fct. congr. extens.
rewrite (fprod_comm FP). reflexivity.
destruct (sig_h x1 x2 H). tauto.
destruct (sig_h x1 x2 H). tauto.
apply unique; assumption.
Qed.

End Types_FProd.

(* --------------------------------------------------------------- *)
(** ** Basic properties of environments.                           *)
(* --------------------------------------------------------------- *)

Lemma env0_push_id : forall X (g : env0 X),
  env0_push (id_bij X) g = g.
Proof.
intros. unfold env0. extens.
unfold env0_push; simpl; intros.
apply fct_id. reflexivity.
Qed.

(* --------------------------------------------------------------- *)
(** ** Some trivial properties of [sub].                           *)
(* --------------------------------------------------------------- *)

Lemma sa_rfl : forall X (x : T X) (g : env0 X), sub g x x.
Proof.
induction x; auto.
Qed.

Hint Resolve sa_rfl.

Lemma sa_tvar : forall (X : Set) (g : env0 X) (x : X),
  sub g (eta x) (g x).
Proof.
auto.
Qed.

Lemma sa_top_inv : forall X g (t : T X), sub g (top X) t -> t = top X.
Proof.
assert (forall X g (s t : T X), s = top X -> sub g s t -> t = top X).
induction 2; try discriminate; auto.
solve [ firstorder ].
Qed.

Lemma sa_eta_inv : forall X (g : env0 X) (x : X) (t : T X),
  sub g t (eta x) -> exists y : X, t = eta y.
Proof.
intros X g.
assert (forall s t : T X, sub g t s ->
        forall x, s = eta x -> exists y : X, t = eta y).
2: eauto.
induction 1; try discriminate 1.
inversion 1.
solve [ intuition eauto ].
solve [ intuition eauto ].
Qed.

(* --------------------------------------------------------------- *)
(** ** Functoriality of [sub].                                     *)
(* --------------------------------------------------------------- *)

Lemma sub_fct :
  forall X (g : env0 X) (m n : T X), sub g m n ->
  forall Y (d : env0 Y) (f : X -> Y),
     (forall x, fct f (g x) = d (f x)) ->
     sub d (fct f m) (fct f n).
Proof.
induction 1; intros; simpl; auto.
apply sa_trans_tvar. rewrite <- H0. auto.
constructor. auto.
apply IHsub2. clear IHsub1 IHsub2.
destruct x; unfold consenv; simpl.
unfold shift. rewrite <- H1. do 2 rewrite fct_fct. congr.
unfold shift. do 2 rewrite fct_fct. reflexivity.
Qed.

Lemma sub_unfct : forall (X Y : Set)(f : X -> Y)
    (d : env0 Y) (g : env0 X) (m n : T X),
  (forall x y, f x = f y -> x = y) ->
  (forall x, fct f (g x) = d (f x)) ->
  (sub d (fct f m) (fct f n)) -> sub g m n.
Proof.
assert
  (forall Y (d : env0 Y) (s t : T Y), (sub d s t) ->
    forall X (g : env0 X) (m n : T X) (f : X -> Y),
     (forall x y, f x = f y -> x = y) ->
     (forall x, fct f (g x) = d (f x)) ->
     s = fct f m -> t = fct f n -> sub g m n).
2: eauto.
induction 1; intros.

assert (n = top _).
apply fct_eq_top with (f := f). auto.
subst n. trivial.

assert (n = m).
2: subst m; auto.
apply fct_inj with (f := f).
assumption.
transitivity (eta x); auto.

subst t.
destruct (@fct_eq_eta _ m _ f x). auto. clear H2.
destruct H3; subst x m.
apply sa_trans_tvar.
apply (IHsub _ g0 (g0 x0) n f); auto.

destruct (@fct_eq_arr _ m _ f s1 s2). auto. clear H3.
destruct H5.
destruct H3 as (Hm, (Hs1, Hs2)).
subst m s1 s2.
destruct (@fct_eq_arr _ n _ f t1 t2). auto. clear H4.
destruct H3.
destruct H3 as (Hn, (Ht1, Ht2)).
subst n t1 t2.
constructor.
eauto.
firstorder idtac.

destruct (@fct_eq_uni _ m _ f s1 s2). auto. clear H3.
destruct H5.
destruct H3 as (Hm, (Hs1, Hs2)).
subst m s1 s2.
destruct (@fct_eq_uni _ n _ f t1 t2). auto. clear H4.
destruct H3.
destruct H3 as (Hn, (Ht1, Ht2)).
subst n t1 t2.
constructor.
eauto.
clear IHsub1.
apply (IHsub2 _ (consenv g0 x1) x0 x2 (optfct f)).
clear IHsub2.
destruct x3; destruct y; simpl; try discriminate 1; auto.
inversion 1.
auto.

destruct x3; unfold consenv, shift; simpl.
rewrite <- H2.
do 2 rewrite fct_fct. congr.
do 2 rewrite fct_fct.
reflexivity.
reflexivity.
reflexivity.
Qed.

Lemma sub_fct_bij : forall X Y (f : Bij X Y) (g : env0 X) x y,
  sub g x y -> sub (env0_push f g) (fct f x) (fct f y).
Proof.
intros.
apply sub_fct with g.
assumption.
intros.
unfold env0_push.
rewrite bij_left.
reflexivity.
Qed.

Lemma sub_unfct_bij : forall X Y (f : Bij X Y) (g : env0 X) x y,
  sub (env0_push f g) (fct f x) (fct f y) -> sub g x y.
Proof.
intros.
apply (@sub_unfct _ _ f (env0_push f g)).
apply bij_inj.
intros.
unfold env0_push.
rewrite bij_left.
reflexivity.
assumption.
Qed.

(** ** Compatibility of [sub] and [shift]. *)

(** The following results are a trival corollary of the
    functoriality of [sub]. *)

Lemma sub_unshift : forall X (g : env0 X) (s t p : T X),
  sub (consenv g p) (shift s) (shift t) -> sub g s t.
Proof.
unfold shift; intros.
eapply sub_unfct with (d := consenv g p) (f := @Some X);
  try reflexivity.
inversion 1; reflexivity.
assumption.
Qed.

Lemma sub_shift : forall X (g : env0 X) s t,
  sub g s t -> forall (p : T X),
  sub (consenv g p) (shift s) (shift t).
Proof.
unfold shift; intros.
apply sub_fct with g; trivial.
Qed.

(* --------------------------------------------------------------- *)
(** ** Compatibility of [sub] with consenv and [appendenv].        *)
(* --------------------------------------------------------------- *)

Lemma sub_consenv_appendenv : forall (X Y : Set)
  (g : env0 X) (e : env X Y) (p : T (X + Y))
  (s t : T (option (X + Y))),
  sub (appendenv g (cenv e p))
    (fct (@opt_right_bij X Y) s) (fct (@opt_right_bij X Y) t) ->
  sub (consenv (appendenv g e) p) s t.
Proof.
intros.
apply sub_unfct_bij with (f := @opt_right_bij X Y).
replace (env0_push (opt_right_bij X Y) (consenv (appendenv g e) p))
  with (appendenv g (cenv e p)).
assumption.
unfold env0. extens x. destruct x; simpl.
unfold env0_push. simpl. unfold consenv. simpl. unfold shift.
do 2 rewrite fct_fct.
congr. simpl. extens. reflexivity.
destruct o. unfold env0_push, cenv. simpl.
unfold consenv, appendenv, shift. rewrite fct_fct. reflexivity.
unfold env0_push. simpl.
unfold cenv, consenv, appendenv, shift. simpl.
rewrite fct_fct. reflexivity.
Qed.

Lemma sub_append_empty : forall (X : Set) (g : env0 X) s t,
  sub (appendenv g (eenv X))
    (fct (inl empty) s) (fct (inl empty) t) ->
  sub g s t.
Proof.
intros.
apply sub_unfct with
    (f := inl (A:=X) empty)
    (d := appendenv g (eenv X));
  trivial.
inversion 1; reflexivity.
Qed.

Lemma sub_append_empty_2 : forall (X : Set) (g : env0 X) s t,
  sub g (fct (@unempty_r X) s) (fct (@unempty_r X) t) ->
  sub (appendenv g (eenv X)) s t.
Proof.
intros.
apply sub_unfct with (f := @unempty_r X) (d := g).
destruct x as [x|x]. 2: destruct x.
destruct y as [y|y]. 2: destruct y.
simpl.
destruct 1; reflexivity.
destruct x as [x|x]. 2: destruct x.
simpl. rewrite fct_fct. simpl. apply fct_id. reflexivity.
assumption.
Qed.
