(* --------------------------------------------------------------- *)
(** * Transitivity and narrowing without well-formedness.          *)
(* --------------------------------------------------------------- *)

Set Implicit Arguments.

Require Export F_sub_facts.

(* --------------------------------------------------------------- *)
(** ** Structure of a type.                                        *)
(* --------------------------------------------------------------- *)

Inductive T' : Set :=
  | eta' : T'
  | top' : T'
  | arr' : T' -> T' -> T'
  | uni' : T' -> T' -> T'.

(* --------------------------------------------------------------- *)
(** Maps a type to its structure.                                  *)
(* --------------------------------------------------------------- *)

Fixpoint tstr (X : Set) (t : T X) {struct t} : T' :=
  match t with
  | eta _ => eta'
  | top => top'
  | arr s t => arr' (tstr s) (tstr t)
  | uni s t => uni' (tstr s) (tstr t)
  end.

Lemma fct_tstr : forall (X : Set) (t : T X) (Y : Set) (f : X -> Y),
  tstr (fct f t) = tstr t.
Proof.
induction t; simpl; auto.
Qed.

(* --------------------------------------------------------------- *)
(** ** The induction proof about transitivity and narrowing.       *)
(* --------------------------------------------------------------- *)

Lemma trans_narrow : forall (q' : T'),
  (forall X (q : T X) (g : env0 X) s t,
     tstr q = q' -> sub g s q -> sub g q t -> sub g s t) /\
  (forall X (q : T X) (g : env0 X) (p : T X)
     Y (e : env (option X) Y) (m n : T (option X + Y)),
     tstr q = q' ->
     sub (appendenv (consenv g q) e) m n ->
     sub g p q -> sub (appendenv (consenv g p) e) m n).
Proof.
assert (HT : forall A B : Prop, A -> (A -> B) -> A /\ B). tauto.
induction q'; intros; apply HT; clear HT;
 destruct q; simpl tstr; try discriminate 1;
 inversion 1; intros.

(* -8: q = eta x; trans *)
clear H.
assert (forall q, sub g s q -> q = eta x -> sub g s t).
2: eauto.
induction 1; try discriminate 1.
inversion 1; subst x0. assumption.
intros; subst t0.
apply sa_trans_tvar.
eapply IHsub; eauto.

(* -7 q = eta x; narrow *)
clear H0.
generalize X g Y e p x m n H1 H2; clear X g Y e p x m n H1 H2.
assert (forall Z (d : env0 Z) a b, sub d a b ->
  forall X Y (f : Bij Z (option X + Y)) g e x p m n,
    fct f a = m -> fct f b = n ->
    (forall u, env0_push f d u = appendenv (consenv g (eta x)) e u) ->
    sub g p (eta x) -> sub (appendenv (consenv g p) e) m n).
2:intros;
  apply (H0 (option X + Y)%type
            (appendenv (consenv g (eta x)) e)
            m n H1 X Y (@id_bij _) g e x);
  solve [ assumption
        |  apply fct_id; reflexivity
        | rewrite env0_push_id; reflexivity ].

induction 1; simpl; intros; subst n m.

(* sa_top *)
trivial.

(* sa_refl_tvar *)
trivial.

(* sa_trans_tvar *)
(* q = eta x0, m = d x, n = t *)
rename X0 into Z; rename g into d.
assert (exists y, x = bij_inv f y).
exists (f x); rewrite bij_left; reflexivity.
destruct H1 as (y,Hy). subst x.
rewrite bij_right.
apply sa_trans_tvar.
unfold env0_push in H3; simpl in H3.
destruct y as [[a|]|b].
(* q in X *)
apply IHsub with (f0 := f) (x := x0); clear IHsub; simpl; trivial.
rewrite H3. reflexivity.
(* q = star *)
apply H with (fct (fun a => inl Y (Some a)) (eta x0)).
reflexivity.
unfold appendenv at 2.
unfold consenv at 2.
unfold shift. rewrite fct_fct.
apply sub_fct with g0. assumption.
simpl; intros. unfold consenv, shift.
rewrite fct_fct. reflexivity.
apply IHsub with (f0 := f) (x := x0); simpl; trivial.
rewrite H3. reflexivity.
(* q in Y *)
apply IHsub with (f0 := f) (x := x0); simpl; trivial.
rewrite H3. reflexivity.

(* -8 : sa_arr *)
constructor; eauto.

(* -7 : sa_uni *)
rename X into Z; rename X0 into X; rename g into d.
constructor.
eauto. clear IHsub1.
apply sub_consenv_appendenv.
apply IHsub2 with
  (f := bij_comp (optfct_bij f) (@opt_right_bij (option X) Y)) (x := x).
rewrite fct_fct. congr. rewrite fct_fct. reflexivity.
unfold env0_push in H2; simpl in H2.
destruct u as [[a|]|[b|]]; simpl;
  unfold env0_push, consenv, shift; simpl.
do 2 rewrite fct_fct.
pose (H2 (inl _ (Some a))).
unfold appendenv, consenv, shift in e0.
rewrite fct_fct in e0.
transitivity
  (fct (fun w => match w with
                 | inl x => inl (option Y) x
                 | inr y => inr _ (Some y)
                 end)
     (fct (fun x : X => inl Y (Some x)) (g0 a))).
rewrite <- e0.
rewrite fct_fct.
reflexivity.
rewrite fct_fct.
reflexivity.
rewrite fct_fct.
simpl optfct.
pose (H2 (inl Y None)).
unfold appendenv, consenv, shift in e0.
rewrite fct_fct in e0.
simpl in e0.
transitivity
  (fct (fun w => match w with
                 | inl x => inl (option Y) x
                 | inr y => inr _ (Some y)
                 end)
    (eta (inl Y (Some x)))).
rewrite <- e0.
rewrite fct_fct. reflexivity.
reflexivity.
unfold cenv, default. rewrite fct_fct.
pose (H2 (inr _ b)).
simpl in e0.
rewrite <- e0. rewrite fct_fct. reflexivity.
unfold consenv, cenv, shift; simpl.
do 2 rewrite fct_fct. reflexivity.
assumption.

(* -6: q = top *)
(* trans *)
rewrite (sa_top_inv H1). trivial.

(* -5: narrow *)
clear H H0.
generalize X g Y e p  m n H1 H2; clear X g Y e p  m n H1 H2.
assert (forall Z (d : env0 Z) a b, sub d a b ->
  forall X Y (f : Bij Z (option X + Y)) g e p m n,
    fct f a = m -> fct f b = n ->
    (forall u, env0_push f d u = appendenv (consenv g (top X)) e u) ->
    sub g p (top X) -> sub (appendenv (consenv g p) e) m n).

2: intros; apply H with
    (d := appendenv (consenv g (top X)) e)
    (f := @id_bij (option X + Y)) (a := m) (b := n);
  solve [ try apply fct_id; try rewrite env0_push_id; auto ].

induction 1; simpl; intros; subst n m.
(* -9: sa_top *)
trivial.

(* -8: sa_refl_tvar *)
trivial.

(* -7: sa_trans_tvar *)
clear H3.
assert (exists y, x = bij_inv f y).
exists (f x).
unfold bij_inv.
simpl.
rewrite bij_l.
reflexivity.
destruct H0.
subst x.
rename x0 into x.
destruct x.
destruct o.
(* -9: x in X *)
apply sa_trans_tvar.
apply IHsub with (f0:=f);
pose (H2 (inl Y (Some x))).
simpl in e0.
unfold consenv, env0_push, shift in e0.
rewrite fct_fct in e0.
simpl in e0.
simpl.
rewrite e0.
rewrite bij_r.
simpl.
unfold consenv, shift.
rewrite fct_fct.
reflexivity.
reflexivity.
assumption.
trivial.
(* -8: x = star *)
pose (H2 (inl Y None)).
simpl in e0.
unfold consenv, env0_push, shift in e0.
simpl in e0.
simpl in H.
rewrite (fct_eq_top _ f e0) in H.
rewrite (sa_top_inv H).
simpl; trivial.

(* -7: x in Y *)
apply sa_trans_tvar.
apply IHsub with (f0:=f).
pose (H2 (inr _ y)).
simpl in e0.
unfold consenv, env0_push, shift in e0.
simpl in e0.
simpl.
rewrite e0.
rewrite bij_r.
simpl.
reflexivity.
reflexivity.
assumption.
trivial.

(* -6: sa_arrow *)
solve [ intuition eauto ].

(* -5: sa_uni *)
clear H4.
rename X into Z. rename X0 into X. rename g0 into d.
constructor.
eauto.
apply sub_consenv_appendenv.
apply IHsub2 with
  (f := bij_comp (optfct_bij f) (@opt_right_bij (option X) Y)).
rewrite fct_fct. congr. rewrite fct_fct. reflexivity.
unfold env0_push in H3; simpl in H3.
destruct u as [[a|]|[b|]]; simpl;
  unfold env0_push, consenv, shift; simpl.
do 2 rewrite fct_fct.
pose (H3 (inl _ (Some a))).
unfold appendenv, consenv, shift in e0.
rewrite fct_fct in e0.
transitivity
  (fct (fun w => match w with
                 | inl x => inl (option Y) x
                 | inr y => inr _ (Some y)
                 end)
     (fct (fun x : X => inl Y (Some x)) (d a))).
rewrite <- e0.
rewrite fct_fct.
reflexivity.
rewrite fct_fct.
reflexivity.
rewrite fct_fct.
simpl optfct.
pose (H3 (inl Y None)).
unfold appendenv, consenv, shift in e0.
rewrite fct_fct in e0.
simpl in e0.
transitivity
  (fct (fun w => match w with
                 | inl x => inl (option Y) x
                 | inr y => inr _ (Some y)
                 end)
    (top (option X + Y))).
rewrite <- e0.
rewrite fct_fct. reflexivity.
reflexivity.
unfold cenv, default. rewrite fct_fct.
pose (H3 (inr _ b)).
simpl in e0.
rewrite <- e0. rewrite fct_fct. reflexivity.
unfold consenv, cenv, shift; simpl.
do 2 rewrite fct_fct. reflexivity.
trivial.

(* -4: q = arr q1 q2; trans *)
inversion H; clear H. subst q'1 q'2.
clear H5 H6.
destruct IHq'1.
assert (Htrans1 : forall (s t : T X),
  sub g s q1 -> sub g q1 t -> sub g s t).
intros; eapply H; eauto. clear H.
assert (Hnarrow1 : forall p Y (e : env (option X) Y)
  (m n : T (option X + Y)),
  sub (appendenv (consenv g q1) e) m n ->
  sub g p q1 -> sub (appendenv (consenv g p) e) m n).
intros; eapply H1; eauto. clear H1.

destruct IHq'2.
assert (Htrans2 : forall (s t : T X),
  sub g s q2 -> sub g q2 t -> sub g s t).
intros; eapply H; eauto. clear H.
assert (Hnarrow2: forall p Y (e : env (option X) Y)
  (m n : T (option X + Y)),
  sub (appendenv (consenv g q2) e) m n ->
  sub g p q2 -> sub (appendenv (consenv g p) e) m n).
intros; eapply H1; eauto. clear H1.

assert (forall a, sub g s a -> a = arr q1 q2 -> sub g s t).
2: intros; eauto. clear H0.
induction 1; try discriminate 1.
intros; subst t0.
apply sa_trans_tvar.
apply IHsub with q1 q2; trivial.
intros Htemp; inversion Htemp; clear Htemp; subst t1 t2.

assert (forall a, sub g a t -> a = arr q1 q2 -> sub g (arr s1 s2) t).
2: eauto.
destruct 1; try discriminate 1.
trivial.
intros Htemp; inversion Htemp; clear Htemp; subst s0 s3.
constructor.
apply Htrans1; auto.
apply Htrans2; auto.

(* -3: q = arr q1 q2; narrow *)
subst q'1 q'2; clear H0.
destruct IHq'1.
assert (Htrans1 : forall (s t : T X),
  sub g s q1 -> sub g q1 t -> sub g s t).
intros; eapply H0; eauto. clear H0.
assert (Hnarrow1 : forall p Y (e : env (option X) Y)
  (m n : T (option X + Y)),
  sub (appendenv (consenv g q1) e) m n ->
  sub g p q1 -> sub (appendenv (consenv g p) e) m n).
intros; eapply H2; eauto. clear H2.
destruct IHq'2.
assert (Htrans2 : forall (s t : T X),
  sub g s q2 -> sub g q2 t -> sub g s t).
intros; eapply H0; eauto. clear H0.
assert (Hnarrow2 : forall p Y (e : env (option X) Y)
  (m n : T (option X + Y)),
  sub (appendenv (consenv g q2) e) m n ->
  sub g p q2 -> sub (appendenv (consenv g p) e) m n).
intros; eapply H2; eauto. clear H2.

generalize X g Y e p q1 q2 m n H1 H4
  H Htrans1 Hnarrow1 Htrans2 Hnarrow2;
clear X g Y e p q1 q2 m n H1 H4
  H Htrans1 Hnarrow1 Htrans2 Hnarrow2.
assert (forall Z (d : env0 Z) a b, sub d a b ->
forall (X Y : Set) (f : Bij Z (option X + Y))
  (g : env0 X) (e : env (option X) Y)
  (p q1 q2 : T X) (m n : T (option X + Y)),
  fct f a = m -> fct f b = n ->
  (forall u, env0_push f d u = appendenv (consenv g (arr q1 q2)) e u) ->
  sub g p (arr q1 q2) ->
  (forall (X0 : Set) (q : T X0) (g0 : env0 X0) (s t : T X0),
    tstr q = arr' (tstr q1) (tstr q2) ->
    sub g0 s q -> sub g0 q t -> sub g0 s t) ->
  (forall s t : T X, sub g s q1 -> sub g q1 t -> sub g s t) ->
  (forall (p0 : T X) (Y0 : Set) (e0 : env (option X) Y0)
    (m0 n0 : T (option X + Y0)),
    sub (appendenv (consenv g q1) e0) m0 n0 ->
    sub g p0 q1 -> sub (appendenv (consenv g p0) e0) m0 n0) ->
  (forall s t : T X, sub g s q2 -> sub g q2 t -> sub g s t) ->
  (forall (p0 : T X) (Y0 : Set) (e0 : env (option X) Y0)
    (m0 n0 : T (option X + Y0)),
    sub (appendenv (consenv g q2) e0) m0 n0 ->
    sub g p0 q2 -> sub (appendenv (consenv g p0) e0) m0 n0) ->
  sub (appendenv (consenv g p) e) m n).
2: intros;
  apply (H (option X + Y)%type
            (appendenv (consenv g (arr q1 q2)) e)
            m n H1 X Y (@id_bij _) g e p q1 q2);
  solve [ assumption
        |  apply fct_id; reflexivity
        | rewrite env0_push_id; reflexivity ].

induction 1; simpl; intros; subst n m.

(* sa_top *)
trivial.

(* sa_refl_tvar *)
trivial.

(* sa_trans_tvar *)
(* q = eta x0, m = d x, n = t *)
rename X0 into Z; rename g0 into d.
assert (exists y, x = bij_inv f y).
exists (f x); rewrite bij_left; reflexivity.
destruct H0 as (y,Hy). subst x.
rewrite bij_right.
apply sa_trans_tvar.
unfold env0_push in H2; simpl in H2.
destruct y as [[a|]|b].
(* q in X *)
apply IHsub with (f0 := f) (q1 := q1) (q2 := q2);
  clear IHsub; simpl; trivial.
rewrite H2. reflexivity.
(* q = star *)
apply H4 with (fct (fun a => inl Y (Some a)) (arr q1 q2)).
rewrite fct_tstr. reflexivity.
unfold appendenv at 2.
unfold consenv at 2.
unfold shift. rewrite fct_fct.
apply sub_fct with d. assumption.
simpl; intros. unfold consenv, shift.
rewrite fct_fct. reflexivity.
apply IHsub with (f0 := f) (q1 := q1) (q2 := q2);
  clear IHsub; simpl; trivial.
rewrite H2.
simpl.
do 2 rewrite fct_fct.
reflexivity.
(* q in Y *)
apply IHsub with (f0 := f) (q1 := q1) (q2 := q2);
  clear IHsub; simpl; trivial.
rewrite H2. reflexivity.

(* sa_arr *)
rename X into Z; rename X0 into X.
rename e into d.
rename g into e.
rename g0 into g.

constructor.
apply IHsub1 with (f := f) (q1 := q1) (q2 := q2); auto.
apply IHsub2 with (f := f) (q1 := q1) (q2 := q2); auto.

(* sa_uni *)
constructor.
apply IHsub1 with (f := f) (q1 := q1) (q2 := q2); auto.
clear IHsub1.
rename X into Z; rename X0 into X; rename g into d.
apply sub_consenv_appendenv.
apply IHsub2 with
  (f := bij_comp (optfct_bij f) (@opt_right_bij (option X) Y))
  (q1 := q1) (q2 := q2); try assumption.
rewrite fct_fct. reflexivity. rewrite fct_fct. reflexivity.
unfold env0_push in H3; simpl in H3.
destruct u as [[a|]|[b|]]; simpl;
  unfold env0_push, consenv, shift; simpl.

do 2 rewrite fct_fct.
pose (H3 (inl _ (Some a))).
unfold appendenv, consenv, shift in e0.
rewrite fct_fct in e0.
transitivity
  (fct (fun w => match w with
                 | inl x => inl (option Y) x
                 | inr y => inr _ (Some y)
                 end)
     (fct (fun x : X => inl Y (Some x)) (g0 a))).
rewrite <- e0. rewrite fct_fct. reflexivity.
rewrite fct_fct.
reflexivity.
do 3 rewrite fct_fct.
pose (H3 (inl Y None)).
unfold appendenv, consenv, shift in e0.
rewrite fct_fct in e0.
simpl in e0.
transitivity
  (fct (fun w => match w with
                 | inl x => inl (option Y) x
                 | inr y => inr _ (Some y)
                 end)
    (arr (fct (fun x : X => inl Y (Some x)) q1)
         (fct (fun x : X => inl Y (Some x)) q2))).
rewrite <- e0.
rewrite fct_fct. reflexivity.
simpl.
do 2 rewrite fct_fct. 
reflexivity.

unfold cenv.
simpl.
rewrite fct_fct.
simpl.
pose (H3 (inr _ b)).
unfold appendenv, consenv, shift in e0.
rewrite <- e0.
rewrite fct_fct.
reflexivity.

unfold cenv.
simpl.
do 2 rewrite fct_fct.
reflexivity.

(* -2: q = uni q1 q2; trans *)
inversion H; clear H. subst q'1 q'2.
clear H5 H6.
destruct IHq'1.
assert (Htrans1 : forall (s t : T X),
  sub g s q1 -> sub g q1 t -> sub g s t).
intros; eapply H; eauto. clear H.
rename H1 into Hnarrow1.

destruct IHq'2.
assert (Htrans2 : forall (g : env0 (option X)) (s t : T (option X)),
  sub g s q2 -> sub g q2 t -> sub g s t).
intros; eapply H; eauto. clear H.
rename H1 into Hnarrow2.

assert (forall a, sub g s a -> a = uni q1 q2 -> sub g s t).
2: intros; eauto. clear H0.
induction 1; try discriminate 1.
intros; subst t0.
apply sa_trans_tvar.
apply IHsub with q1 q2; trivial.
intros Htemp; inversion Htemp; clear Htemp; subst t1 t2.

assert (forall a, sub g a t -> a = uni q1 q2 -> sub g (uni s1 s2) t).
2: eauto.
destruct 1; try discriminate 1.
trivial.
intros Htemp; inversion Htemp; clear Htemp; subst s0 s3.
constructor.
apply Htrans1; auto.
apply Htrans2; auto.
apply sub_append_empty.
apply Hnarrow1 with (q := q1).
reflexivity.
apply sub_append_empty_2.
do 2 rewrite fct_fct.
rewrite fct_id.
rewrite fct_id.
assumption.
destruct u; simpl; reflexivity.
destruct u; simpl; reflexivity.
assumption.

(* -1: q = uni q1 q2; narrow *)
subst q'1 q'2; clear H0.
destruct IHq'1.
assert (Htrans1 : forall (s t : T X),
  sub g s q1 -> sub g q1 t -> sub g s t).
intros; eapply H0; eauto. clear H0.
assert (Hnarrow1 : forall p Y (e : env (option X) Y)
  (m n : T (option X + Y)),
  sub (appendenv (consenv g q1) e) m n ->
  sub g p q1 -> sub (appendenv (consenv g p) e) m n).
intros; eapply H2; eauto. clear H2.
destruct IHq'2.
assert (Htrans2 : forall (g : env0 (option X)) (s t : T (option X)),
  sub g s q2 -> sub g q2 t -> sub g s t).
intros; eapply H0; eauto. clear H0.
assert (Hnarrow2 :
  forall p Y (g : env0 (option X)) (e : env (option (option X)) Y)
  (m n : T (option (option X) + Y)),
  sub (appendenv (consenv g q2) e) m n ->
  sub g p q2 -> sub (appendenv (consenv g p) e) m n).
intros; eapply H2; eauto. clear H2.

generalize X g Y e p q1 q2 m n H1 H4
  H Htrans1 Hnarrow1 Htrans2 Hnarrow2;
clear X g Y e p q1 q2 m n H1 H4
  H Htrans1 Hnarrow1 Htrans2 Hnarrow2.
assert (forall Z (d : env0 Z) a b, sub d a b ->
forall (X Y : Set) (f : Bij Z (option X + Y))
  (g : env0 X) (e : env (option X) Y)
  (p q1 : T X) q2 (m n : T (option X + Y)),
  fct f a = m -> fct f b = n ->
  (forall u, env0_push f d u = appendenv (consenv g (uni q1 q2)) e u) ->
  sub g p (uni q1 q2) ->
  (forall (X0 : Set) (q : T X0) (g0 : env0 X0) (s t : T X0),
    tstr q = uni' (tstr q1) (tstr q2) ->
    sub g0 s q -> sub g0 q t -> sub g0 s t) ->
  (forall s t : T X, sub g s q1 -> sub g q1 t -> sub g s t) ->
  (forall (p0 : T X) (Y0 : Set) (e0 : env (option X) Y0)
    (m0 n0 : T (option X + Y0)),
    sub (appendenv (consenv g q1) e0) m0 n0 ->
    sub g p0 q1 -> sub (appendenv (consenv g p0) e0) m0 n0) ->
  (forall (g : env0 (option X)) (s t : T (option X)),
    sub g s q2 -> sub g q2 t -> sub g s t) ->
  (forall (g : env0 (option X)) (p0 : T (option X)) (Y0 : Set)
    (e0 : env (option (option X)) Y0)
    (m0 n0 : T (option (option X) + Y0)),
    sub (appendenv (consenv g q2) e0) m0 n0 ->
    sub g p0 q2 -> sub (appendenv (consenv g p0) e0) m0 n0) ->
  sub (appendenv (consenv g p) e) m n).
2: intros;
  apply (H (option X + Y)%type
            (appendenv (consenv g (uni q1 q2)) e)
            m n H1 X Y (@id_bij _) g e p q1 q2);
  try solve [ assumption
        |  apply fct_id; reflexivity
        | rewrite env0_push_id; reflexivity ].
2: auto.
induction 1; simpl; intros; subst n m.

(* sa_top *)
trivial.

(* sa_refl_tvar *)
trivial.

(* sa_trans_tvar *)
(* q = uni q1 q2, m = d x, n = t *)
rename X into Z; rename X0 into X; rename g0 into d.
assert (exists y, x = bij_inv f y).
exists (f x); rewrite bij_left; reflexivity.
destruct H0 as (y,Hy). subst x.
rewrite bij_right.
apply sa_trans_tvar.
unfold env0_push in H2; simpl in H2.
destruct y as [[a|]|b].
(* q in X *)
apply IHsub with (f0 := f) (q1 := q1) (q2 := q2);
  clear IHsub; simpl; trivial.
rewrite H2. reflexivity.
(* q = star *)
apply H4 with (fct (fun a => inl Y (Some a)) (uni q1 q2)).
rewrite fct_tstr. reflexivity.
unfold appendenv at 2.
unfold consenv at 2.
unfold shift. rewrite fct_fct.
apply sub_fct with d. assumption.
simpl; intros. unfold consenv, shift.
rewrite fct_fct. reflexivity.
apply IHsub with (f0 := f) (q1 := q1) (q2 := q2);
  clear IHsub; simpl; trivial.
rewrite H2.
simpl.
do 2 rewrite fct_fct.
congr. congr. extens x. destruct x; reflexivity.
(* q in Y *)
apply IHsub with (f0 := f) (q1 := q1) (q2 := q2);
  clear IHsub; simpl; trivial.
rewrite H2. reflexivity.

(* sa_arr *)
rename X into Z; rename X0 into X.
rename e into d.
rename g into e.
rename g0 into g.

constructor.
apply IHsub1 with (f := f) (q1 := q1) (q2 := q2); auto.
apply IHsub2 with (f := f) (q1 := q1) (q2 := q2); auto.

(* sa_uni *)
constructor.
apply IHsub1 with (f := f) (q1 := q1) (q2 := q2); auto.
clear IHsub1.
rename X into Z; rename X0 into X; rename g into d.
apply sub_consenv_appendenv.
apply IHsub2 with
  (f := bij_comp (optfct_bij f) (@opt_right_bij (option X) Y))
  (q1 := q1) (q2 := q2); try assumption.
rewrite fct_fct. reflexivity. rewrite fct_fct. reflexivity.
unfold env0_push in H3; simpl in H3.
destruct u as [[a|]|[b|]]; simpl;
  unfold env0_push, consenv, shift; simpl.

do 2 rewrite fct_fct.
pose (H3 (inl _ (Some a))).
unfold appendenv, consenv, shift in e0.
rewrite fct_fct in e0.
transitivity
  (fct (fun w => match w with
                 | inl x => inl (option Y) x
                 | inr y => inr _ (Some y)
                 end)
     (fct (fun x : X => inl Y (Some x)) (g0 a))).
rewrite <- e0. rewrite fct_fct. reflexivity.
rewrite fct_fct.
reflexivity.
do 3 rewrite fct_fct.
pose (H3 (inl Y None)).
unfold appendenv, consenv, shift in e0.
rewrite fct_fct in e0.
simpl in e0.
transitivity
  (fct (fun w => match w with
                 | inl x => inl (option Y) x
                 | inr y => inr _ (Some y)
                 end)
    (uni (fct (fun x : X => inl Y (Some x)) q1)
         (fct (optfct (fun x : X => inl _ (Some x))) q2))).
rewrite <- e0.
rewrite fct_fct. reflexivity.
simpl. do 2 rewrite fct_fct. 
congr. congr. extens x. destruct x; reflexivity.

unfold cenv. simpl.
rewrite fct_fct. simpl.
pose (H3 (inr _ b)).
unfold appendenv, consenv, shift in e0.
rewrite <- e0. rewrite fct_fct. reflexivity.

unfold cenv. simpl.
do 2 rewrite fct_fct. reflexivity.
Qed.

(* --------------------------------------------------------------- *)
(** ** Narrowing for subtyping without well-formedness.            *)
(* --------------------------------------------------------------- *)

Lemma fsub_narrow : forall X (q : T X) (g : env0 X) (p : T X)
    Y (e : env (option X) Y) (m n : T (option X + Y)),
  sub (appendenv (consenv g q) e) m n ->
  sub g p q -> sub (appendenv (consenv g p) e) m n.
Proof.
intros.
assert (TN := trans_narrow (tstr q)).
destruct TN. eauto.
Qed.

(* --------------------------------------------------------------- *)
(** ** Transitivity for subtyping without well-formedness.         *)
(* --------------------------------------------------------------- *)

Lemma fsub_trans : forall X (q : T X) (g : env0 X) s t,
  sub g s q -> sub g q t -> sub g s t.
Proof.
intros.
assert (TN := trans_narrow (tstr q)).
destruct TN. eauto.
Qed.
