(* --------------------------------------------------------------- *)
(** * Narrowing and Transitivity for well formed subtyping.        *)
(* --------------------------------------------------------------- *)

Set Implicit Arguments.

Require Export F_sub_trans.
Require Export Well_formed_env.

(* --------------------------------------------------------------- *)
(** Definition of subtyping.                                       *)
(* --------------------------------------------------------------- *)

Definition WFsub (X : Set) (g : env0 X) (s t : T X) :=
  WF0all g /\ sub g s t.

(* --------------------------------------------------------------- *)
(** Narrowing for subtyping.                                       *)
(* --------------------------------------------------------------- *)

Theorem WFsub_narrow : forall X (q : T X) (g : env0 X) (p : T X)
    Y (e : env (option X) Y) (m n : T (option X + Y)),
  WFsub (appendenv (consenv g q) e) m n ->
  WFsub g p q -> WFsub (appendenv (consenv g p) e) m n.
Proof.
destruct 1. destruct 1. unfold WFsub.
split.
apply WF0all_appendenv.
apply WF0all_consenv. assumption.
apply (appendenv_WF0all_2 H).
assert (HN := fsub_narrow). eauto.
Qed.

(* --------------------------------------------------------------- *)
(** Transitivity for subtyping.                                    *)
(* --------------------------------------------------------------- *)

Theorem WFsub_trans : forall X (q : T X) (g : env0 X) s t,
  WFsub g s q -> WFsub g q t -> WFsub g s t.
Proof.
destruct 1. destruct 1. unfold WFsub.
assert (HT := fsub_trans). eauto.
Qed.
